# Accepts a path and recursively reports all unique NTFS permissions for all child folders and files
# Shaw; Home; NTFS; Permissions; Module; Report;
# $Revision: 368 $
#requires -version 3

Function Get-UniqueNTFSPermissions
{
  <#
.SYNOPSIS
    Accepts a path and recursively reports all unique NTFS permissions for all child folders and files
.DESCRIPTION
    This function also optionally accepts the ability to automatically remove a list of known NTFS permissions from the results as well.
    Ignores SIDs by default
	Note: On a SSD local disk scanning C:\Windows, this script processes roughly 120 files/folders per second or roughly 430,000 files/folders per hour
	
.PARAMETER Path
    The root path you'd like to scan.  Note can also be a UNC path
.PARAMETER ExcludedAccounts
    Specifes a commma seperated list of user accounts to exclude from the resulting report
.EXAMPLE
    Get-UniqueNTFSPermissions
    This will prompt for a path to scan
.EXAMPLE
    Get-UniqueNTFSPermissions D:\Data
    This will scan the D:\drive and will return 4 columns: The path of the unique file or folder, the current filtered permissions, added permissions, removed permissions

.EXAMPLE
    Get-UniqueNTFSPermissions D:\Data -ExcludedAccounts "BUILTIN\Administrators, BUILTIN\Users, NT AUTHORITY\SYSTEM, CREATOR OWNER"
    This will scan the D:\drive and will automatically exclude the 4 common accounts from report

.EXAMPLE
    Get-UniqueNTFSPermissions D:\Data -ExcludedAccounts "NT AUTHORITY\SYSTEM, CREATOR OWNER" -IncludeSIDs
    This will scan the D:\drive and will automatically exclude the 2 common accounts from report but will also include any unknown SIDs

.EXAMPLE
    $ExcludedAccounts = @"
    BUILTIN\Administrators
    CREATOR OWNER
    NT AUTHORITY\SYSTEM
    NT SERVICE\TrustedInstaller
    "@ -split "`n" -join ","

    Get-UniqueNTFSPermissions d:\data -ExcludedAccounts $ExcludedAccounts -IncludeSIDs | ConvertTo-CSV -Delimiter * -NoTypeInformation | clip
    
    This will scan the D:\drive and will automatically exclude the list of defined common accounts from the report but will also include any unknown SIDs
    It will then export it to the clipboard for pasting into Excel.  Using the "*" delimiter as the default comma is used to seperate the permissions

.NOTES
    Author: Robert Vance
    Date:   November 30th, 2014
.LINK
    http://pleasework.robbievance.net    
#>
  
  [cmdletbinding()]
    Param(
        [parameter(mandatory=$true,position=0, HelpMessage="Provide the path to the root folder you'd like to report permissions on, eg: ('D:\Data' or '\\Server\Data')")]
        [ValidateScript({Test-Path $_})] 
        [string]$Path,
		[parameter(mandatory=$false,position=1)]
        $ExcludedAccounts,
        [parameter(mandatory=$false,position=2)]
        [switch]$IncludeSIDs
    )

    # Paramater accepts comma delimited list of accounts to exclude.  
    # We first split it so that it is a collection that we can use compare-object with
    # We then trim it in case the user entered a space after the comma when entering the names
    $ExcludedAccounts = $ExcludedAccounts -split ","
    $ExcludedAccounts = $ExcludedAccounts.trim()

    # Create a hash table that will store all of the recursive folders, files and NTFS permissions of the chosen folder
    $PermHash = [ordered]@{}
	$Script:OutputObject = $null
   	
    # As this can be a potentially very long running operation, include the the start time in the progress bar
    $StartTime = Get-Date -Format "hh:mm:ss tt"

    If ($PSBoundParameters['Debug']) {  $DebugPreference = 'Continue' }

    try
    {	
        # As Get-ChildItem does not return the parent folder, we have to manually grab the permissions for the parent folder first
        # This also means we have to call Get-Permissions with two different paths which is why we've made it a function
        # It's also why we have to effectively the paramaters from Get-UniqueNTFSPermissions to Get-Permissions
        $RootPerms = Get-Permissions $Path $ExcludedAccounts $IncludeSIds
		$PermHash.Add([string]$Path,$RootPerms)

	    # Loop through each folder and file under the original folder provided and build a hash table including the folder/file and the effective permissions
        Get-ChildItem $Path -Recurse -ErrorVariable Err -ErrorAction SilentlyContinue | ForEach-Object {
   			Write-Progress -Activity "(Task 1 of 2) - Reading Files and Folders... [Started: $StartTime]" -Status $_.FullName
            $Perms = Get-Permissions $_.FullName $ExcludedAccounts $IncludeSIds
            if($Perms) { $PermHash.Add($_.FullName,$Perms) }
	    }
            # Once all folders have been processed, manually add any folders that generated errors to our output object 
            if($Err) { $OutputObject+=@([pscustomobject] @{Path=$Err.TargetObject;Permission=$Err.Exception.Message;Added="";Removed=""}) }
    
        # Now that we have a list of all files and folders, take each one and compare the permissions on that item to that of the parent
        # If the permissions match, don't write it out.  If they are different, write the file or folder to our output object
        $PermHash.GetEnumerator() | % { 
    	    Write-Progress -Activity "(Task 2 of 2) Reading Hashes... [Started: $StartTime]" -Status $($_.key)

            # Give the current permission in the hash table a more friendly name to make it easier to understand what's happening
            $CurrPerm = $($_.value)

            # We want to compare the permissions on each file/folder to its parent.  If they match, ignore it.  If they don't include it
            # This determines what the parent folder is based on the full file/folder path
            $ParentDir = Split-Path $($_.key) -Parent 

            Write-Debug "ParentDir: $ParentDir     |     Path: $Path       |      CurrentPath: $($_.key)"
    		# If this is root folder, can't go up any further so provide the permissions for the same folder
	    	if($ParentDir -eq $Path) { $ParentPerm = ($PermHash.Get_Item(($Path))) }
		    Else
            {
                # Perform a lookup to determine what permission the parent folder has for this item
                $ParentPerm = ($PermHash.Get_Item(($ParentDir))) 
            }
                
            if($ParentPerm) 
            { 
                    $Removed = (Compare-Object $ParentPerm $CurrPerm | where {$_.SideIndicator -eq "<="}).InputObject
                }
            

            # Join objects together into a single line otherwise they'll show up in the report as "[System.Object]"
            $ParentPerm = $ParentPerm -join ", "
            $CurrPerm = $CurrPerm -join ", "
            $Added = $Added -join ", "
            $Removed = $Removed -join ", "

            # Check if the current file or folders permissions match that of the parent.  If they don't, include it in the output object
			if($ParentPerm -ne $CurrPerm)
            {
				$OutputObject+=@([pscustomobject] @{Path=$($_.key);Permission=$CurrPerm;Added=$Added;Removed=$Removed}) 
            }
  
        }
    }    

    # Write any error messages that might be encountered
    catch { Write-Host $_ -ForegroundColor Red }
   	
    $OutputObject = $OutputObject | sort Path
    return $OutputObject
   
}

Function Get-Permissions($Path, $ExcludedAccounts, [bool]$IncludeSIDs)
{ 
    try 
    {
       # Our compare-object code must be an exact match so we can't use it to perform a wildcard lookup for any SID
       # That's why we do the SID lookup here depending on if the parameter was requested or not
       switch($IncludeSIDs)
        { 
            $true {$FullPerms = ((Get-Acl -LiteralPath $Path | select -expandproperty access | select identityreference -unique).identityreference) }
            $false {$FullPerms = ((Get-Acl -LiteralPath $Path | select -expandproperty access | select identityreference -unique).identityreference) | Where-Object { $_ -notmatch 'S-1-5' } }
        }

        if($FullPerms) { $EffectivePerms = ((Compare-Object $FullPerms.Value $ExcludedAccounts | where Sideindicator -eq "<=").InputObject | sort) }

        return $EffectivePerms
    }
    catch [UnauthorizedAccessException]
    {
       	if($_.TargetObject) { $OutputObject+=@([pscustomobject] @{Path=$($_.TargetObject);Permission=$_.Exception;Added="";Removed=""}) }
    }

}